from __future__ import annotations

from dis import Instruction, get_instructions
from types import CodeType
from typing import Sequence

from dismod.utils import neighborhood, search_key_in_iterable


def _extract_import_instructions(
    instructions: list[Instruction],
) -> list[Instruction]:
    """ """
    import_instructions = []
    for instruction in instructions:
        if instruction.opcode in (84, 108, 109):
            import_instructions.append(instruction)

    return import_instructions


def _convert_codeblocks_instructions(
    instructions: list[Instruction],
) -> list[Instruction]:
    """ """
    converted_instructions = []
    for instruction in instructions:
        if instruction.opcode == 100 and isinstance(
            instruction.argval,
            CodeType,
        ):
            extracted_instructions = _extract_import_instructions(
                list(get_instructions(instruction.argval)),
            )
            converted_instructions.extend(extracted_instructions)

    return converted_instructions


def get_instructions_from_file(filepath: str) -> list[Instruction]:
    contents = None
    with open(filepath) as file:
        contents = file.read()

    return list(get_instructions(contents))  # type: ignore


def get_import_instructions(
    instructions: list[Instruction],
) -> list[Instruction]:
    normalized_codeblocks = _convert_codeblocks_instructions(
        instructions=instructions,
    )
    normalized_codeblocks.extend(instructions)
    return _extract_import_instructions(instructions=normalized_codeblocks)


def _check_from_statement(current: int, following: int) -> bool:
    """ """
    return (current == 108 and following == 109) or (
        current == 108 and following == 84
    )


def _check_import_from_statement(
    previous: int,
    current: int,
    following: int | None,
) -> bool:
    """ """
    return (
        (current == 109 and following == 109)
        or (current == 84 and previous == 108)
        or (current == 109 and previous == 108)
    ) or (current == 109 and following is None)


def _check_single_import_statement(current: int, following: int) -> bool:
    """ """
    return (current == 108 and following == 108) or (
        current == 108 and following is None
    )


def parse_instructions(
    instructions: list[Instruction],
) -> list[dict[str, Sequence[str] | None]]:
    """ """
    current_from_statement_name = ""
    parsed_instructions: list[dict[str, Sequence[str] | None]] = []
    for previous, current, following in neighborhood(iterable=instructions):
        # Try to handle from statements import name
        if following and _check_from_statement(
            current=current.opcode,
            following=following.opcode,
        ):
            # Associate the argval from the current instruction being analyzed
            # If we find a '' in the argval property, it means that we are
            # handling a from . import <something>
            current_from_statement_name = (
                current.argval if current.argval else "."
            )

            _, has_from_statement = search_key_in_iterable(
                key="from_statement",
                value=current_from_statement_name,
                iterable=parsed_instructions,
            )
            if not has_from_statement:
                # Initialize a basic structure for the from/import statements
                parsed_instructions.append(
                    {
                        "from_statement": current_from_statement_name,
                        "import_statements": [],
                    },
                )

        # Try to handle import statements that are derivade from a
        # from statement
        if _check_import_from_statement(
            previous=previous.opcode if previous else None,
            current=current.opcode,
            following=following.opcode if following else None,
        ):
            index, _ = search_key_in_iterable(
                key="from_statement",
                value=current_from_statement_name,
                iterable=parsed_instructions,
            )
            parsed_instructions[index]["import_statements"].append(
                current.argval if current.argval else "*",
            )

        # Handle single import statements, no matter where they are
        if _check_single_import_statement(
            current=current.opcode,
            following=following.opcode if following else None,
        ):
            parsed_instructions.append(
                {
                    "from_statement": None,
                    "import_statements": [current.argval],
                },
            )

    return parsed_instructions



from __future__ import annotations

from typing import Any


class DependencyContainer:
    def __init__(self, filepath: str) -> None:
        self.filepath = filepath
        self.basename: str = self._parse_basename()
        self.imports: list[Any] = []

    def _parse_basename(self) -> str:
        return self.filepath.replace("/", ".").replace("\\", ".")

    def add_import(
        self,
        imports: list[Any],
    ) -> None:
        self.imports = imports



from __future__ import annotations

import argparse

from dismod.dependency import DependencyContainer
from dismod.dot import render
from dismod.instruction import (
    get_import_instructions,
    get_instructions_from_file,
    parse_instructions,
)
from dismod.utils import collect_files_in_module


def create_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="dismod",
        description="Generate dependency graphs for your codebase.",
    )

    parser.add_argument(
        "filepath",
        metavar="filepath",
        type=str,
        help="The path to the source code in your project.",
    )
    parser.add_argument(
        "--split-files",
        action="store_true",
        default=False,
        help="Render the output as split files",
    )
    parser.add_argument("--ignore-folder", help="Ignore a specific folder")
    parser.add_argument(
        "--rankdir",
        default="LR",
        help="Sets the direction of the graph. See also: https://graphviz.org/docs/attrs/ratio/",  # noqa: E501
    )
    parser.add_argument(
        "--overlap",
        default="scale",
        help="Determine if and how node overlaps should be removed. See also: https://graphviz.org/docs/attrs/overlap/",  # noqa: E501
    )
    parser.add_argument(
        "--ratio",
        default="fill",
        help="Sets the aspect ratio (drawing height/drawing width) for the drawing. See also: https://graphviz.org/docs/attrs/ratio/",  # noqa: E501
    )
    parser.add_argument(
        "--fontsize",
        default="16",
        help="The size of the font to be used. See also: https://graphviz.org/docs/attrs/fontsize/",  # noqa: E501
    )
    parser.add_argument(
        "--dpi",
        default="150",
        help="Specifies the expected number of pixels per inch on a display device. See also: https://graphviz.org/docs/attrs/dpi/",  # noqa: E501
    )
    parser.add_argument(
        "--shape",
        default="rectangle",
        help="Sets the shape of a node. See also: https://graphviz.org/docs/attrs/shape/",  # noqa: E501
    )
    parser.add_argument(
        "--fontname",
        default="Consolas",
        help="Font used for text. See also: https://graphviz.org/docs/attrs/fontname/",  # noqa: E501
    )
    return parser


def get_import_dependency_list(files: list[str]) -> list[DependencyContainer]:
    import_dependency_list = []

    for file in files:
        instructions = get_instructions_from_file(filepath=file)
        all_instructions = get_import_instructions(instructions=instructions)
        # Early exit, we don't care about files that don't have any import
        # statements, e.g; __init__.py or other simple file that has no import
        if not all_instructions:
            continue

        instructions_normalized = parse_instructions(
            instructions=all_instructions,
        )

        file_import_dependencies = DependencyContainer(filepath=file)
        file_import_dependencies.add_import(instructions_normalized)
        import_dependency_list.append(file_import_dependencies)
    return import_dependency_list


def main() -> int:
    parser = create_argument_parser()
    args = parser.parse_args()

    # Collect all files that lives in the specified filepath
    files = collect_files_in_module(
        filepath=args.filepath,
        ignore_folder=args.ignore_folder,
    )
    # Get a list of import dependencies for each file
    list_of_dependencies = get_import_dependency_list(files=files)

    render(
        project_name=args.filepath,
        dependency_containers=list_of_dependencies,
        rankdir=args.rankdir,
        overlap=args.overlap,
        ratio=args.ratio,
        fontsize=args.fontsize,
        dpi=args.dpi,
        shape=args.shape,
        fontname=args.fontname,
        split_files=args.split_files,
    )

    return 0



from __future__ import annotations

import os

from dismod.dependency import DependencyContainer

DOT_TEMPLATE = """\
digraph "%s" {
    graph [
        rankdir="%s",
        overlap="%s",
        ratio="%s",
        fontsize="%s",
        dpi="%s",
    ];
    node [
        fontsize="%s"
        shape="%s"
        fontname="%s"
    ];

    %s}
"""


def render(
    project_name: str,
    dependency_containers: list[DependencyContainer],
    rankdir: str,
    overlap: str,
    ratio: str,
    fontsize: str,
    dpi: str,
    shape: str,
    fontname: str,
    split_files: bool,
) -> None:
    if not os.path.exists("renders"):
        os.mkdir("renders")

    if split_files:
        _render_multiple_files(
            project_name=project_name,
            dependency_containers=dependency_containers,
            rankdir=rankdir,
            overlap=overlap,
            ratio=ratio,
            fontsize=fontsize,
            dpi=dpi,
            shape=shape,
            fontname=fontname,
        )
    else:
        _render_cluster_files(
            project_name=project_name,
            dependency_containers=dependency_containers,
            rankdir=rankdir,
            overlap=overlap,
            ratio=ratio,
            fontsize=fontsize,
            dpi=dpi,
            shape=shape,
            fontname=fontname,
        )


def _render_cluster_files(
    project_name: str,
    dependency_containers: list[DependencyContainer],
    rankdir: str,
    overlap: str,
    ratio: str,
    fontsize: str,
    dpi: str,
    shape: str,
    fontname: str,
) -> None:
    # Dirty hack to prevent a \t in the first edge
    edges = """
"""

    for dependency in dependency_containers:
        for _import in dependency.imports:
            basename = dependency.basename.rsplit(".", 1)[0]
            if _import["from_statement"] is not None:
                edges += f'\t"{basename}" -> "{_import["from_statement"]}";\n'

            from_statement = (
                _import["from_statement"]
                if _import["from_statement"]
                else basename
            )

            edges += '\t"{}" -> {{"{}"}};\n'.format(
                from_statement,
                '","'.join(_import["import_statements"]),
            )

    with open(f"renders/{project_name}.dot", mode="w") as file:
        file.write(
            DOT_TEMPLATE
            % (
                project_name,
                rankdir,
                overlap,
                ratio,
                fontsize,
                dpi,
                fontsize,
                shape,
                fontname,
                edges,
            ),
        )


def _render_multiple_files(
    project_name: str,
    dependency_containers: list[DependencyContainer],
    rankdir: str,
    overlap: str,
    ratio: str,
    fontsize: str,
    dpi: str,
    shape: str,
    fontname: str,
) -> None:
    for dependency in dependency_containers:
        # Dirty hack to prevent a \t in the first edge
        edges = """
"""
        for _import in dependency.imports:
            basename = dependency.basename.rsplit(".", 1)[0]
            if _import["from_statement"] is not None:
                edges += f'\t"{basename}" -> "{_import["from_statement"]}";\n'

            from_statement = (
                _import["from_statement"]
                if _import["from_statement"]
                else basename
            )

            edges += '\t"{}" -> {{"{}"}};\n'.format(
                from_statement,
                '", "'.join(_import["import_statements"]),
            )

        with open(f"renders/{basename}.dot", mode="w") as file:
            file.write(
                DOT_TEMPLATE
                % (
                    project_name,
                    rankdir,
                    overlap,
                    ratio,
                    fontsize,
                    dpi,
                    fontsize,
                    shape,
                    fontname,
                    edges,
                ),
            )



from __future__ import annotations

import fnmatch
import os
from typing import Any, Generator


def collect_files_in_module(filepath: str, ignore_folder: str) -> list[str]:
    """ """
    matches = []
    for root, dirs, filenames in os.walk(filepath):
        [
            dirs.remove(d)  # type: ignore
            for d in list(dirs)
            if ignore_folder and (d in ignore_folder)
        ]
        for filename in fnmatch.filter(filenames, "*.py"):
            matches.append(os.path.join(root, filename))

    return matches


def neighborhood(
    iterable: list[Any],
) -> Generator[Any, Any, Any]:
    """ """
    if not iterable:
        return iterable

    iterator = iter(iterable)
    prev_item = None
    current_item = next(iterator)  # throws StopIteration if empty.
    for next_item in iterator:
        yield (prev_item, current_item, next_item)
        prev_item = current_item
        current_item = next_item
    yield (prev_item, current_item, None)


def search_key_in_iterable(
    key: str,
    value: str,
    iterable: list[Any],
) -> tuple[Any, ...]:
    """ """
    return tuple(
        next(
            (
                (index, thing)
                for index, thing in enumerate(iterable)
                if thing[key] == value
            ),
            (None, None),
        ),
    )


def split_list_in_chunks(
    elements: list[Any],
    chunk_size: int,
) -> Generator[Any, Any, Any]:
    """Yield chunk_size number of striped chunks from elements."""
    for index in range(0, chunk_size):
        yield elements[index::chunk_size]
